from __future__ import annotations

import importlib
import pkgutil
import string
import sys


def b62encode(data: bytes) -> str:
    num = int.from_bytes(data, "big")
    characters = string.digits + string.ascii_letters

    encoded = ""
    base: int = len(characters)

    if num < 0:
        return ""

    while num >= base:
        mod = num % base
        num //= base
        encoded = characters[mod] + encoded

    if num > 0:
        encoded = characters[num] + encoded

    return encoded


def import_submodules(package_name: str) -> None:
    importlib.invalidate_caches()

    sys.path.append(".")

    # Import at top level
    module = importlib.import_module(package_name)
    path = getattr(module, "__path__", [])
    path_string = "" if not path else path[0]

    for module_finder, name, _ in pkgutil.walk_packages(path):
        if path_string and getattr(module_finder, "path") != path_string:  # noqa: B009
            continue
        subpackage = f"{package_name}.{name}"
        import_submodules(subpackage)


def import_modules(module_names: list[str]) -> None:
    for module_name in module_names:
        import_submodules(module_name)



from __future__ import annotations

import os
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, Iterator

from cachestore.common import FileLock
from cachestore.storages.storage import Storage


class LocalStorage(Storage):
    def __init__(self, root: str | PathLike | None = None) -> None:
        self._root = Path(root or Path.cwd()).absolute()

        self._root.mkdir(parents=True, exist_ok=True)

    def __str__(self) -> str:
        return f"LocalStorage(root={self._root.relative_to(Path.cwd())})"

    @contextmanager
    def open(self, key: str, mode: str) -> Iterator[IO[Any]]:
        filename = self._root / key
        lockfile = filename.parent / (filename.name + ".lock")
        filename.parent.mkdir(parents=True, exist_ok=True)
        with FileLock(lockfile):
            try:
                with open(filename, mode) as fp:
                    yield fp
            except (Exception, KeyboardInterrupt):
                os.remove(filename)
                raise

    def remove(self, key: str) -> None:
        filename = self._root / key
        os.remove(filename)

    def exists(self, key: str) -> bool:
        return Path(key).exists()

    def all(self) -> Iterator[str]:
        for filename in self._root.glob("*"):
            if not filename.name.endswith(".lock"):
                yield filename.name



import hashlib
import io
import pickle
from typing import Any

from cachestore.hashers.hasher import Hasher
from cachestore.util import b62encode


class PickleHasher(Hasher):
    def __call__(self, obj: Any) -> str:
        m = hashlib.blake2b()
        with io.BytesIO() as buf:
            pickle.dump(obj, buf)
            m.update(buf.getbuffer())
            return b62encode(m.digest())



import argparse
import importlib

from cachestore.commands.subcommand import Subcommand
from cachestore.common import Selector
from cachestore.util import import_modules


@Subcommand.register("remove")
class RemoveCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument("cache")
        self.parser.add_argument("-f", "--function", action="append", default=[])
        self.parser.add_argument("-a", "--all", action="store_true")
        self.parser.add_argument("--include-package", action="append", default=[])

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        cache_module, cache_name = args.cache.split(":", 1)
        cache = getattr(importlib.import_module(cache_module), cache_name)

        funcinfos = {funcinfo.name: funcinfo for funcinfo in cache.funcinfos() if cache.exists(funcinfo)}
        if args.all:
            funcnames = set(funcinfos.keys())
        elif args.function:
            funcnames = set(args.function)
        else:
            funcnames = set()
            candidates = [name for name, info in funcinfos.items()]
            if candidates:
                selected_funcname = Selector()(candidates)
                funcnames = set() if selected_funcname is None else {selected_funcname}

        if funcnames:
            for funcname in funcnames:
                if args.all or funcname in funcinfos:
                    print(f"remove: {funcname}")
                    cache.remove(funcinfos[funcname])
                else:
                    print(f"skip  : {funcname}")
        else:
            print("No caches to remove.")



import abc
from contextlib import contextmanager
from typing import IO, Any, Iterator


class Storage(abc.ABC):
    @abc.abstractmethod
    @contextmanager
    def open(self, key: str, mode: str) -> Iterator[IO[Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, key: str) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, key: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def all(self) -> Iterator[str]:
        raise NotImplementedError



from __future__ import annotations

from functools import wraps
from logging import getLogger
from typing import Any, Callable, TypeVar, cast

from cachestore.formatters import Formatter, PickleFormatter
from cachestore.hashers import Hasher, PickleHasher
from cachestore.metadata import ExecutionInfo, FunctionInfo
from cachestore.storages import LocalStorage, Storage

DEFAULT_CACHE_DIR = ".cache"

logger = getLogger(__name__)

T = TypeVar("T")


class Cache:
    def __init__(
        self,
        storage: Storage | None = None,
        formatter: Formatter | None = None,
        hasher: Hasher | None = None,
    ) -> None:
        self.storage = storage or LocalStorage(DEFAULT_CACHE_DIR)
        self.formatter = formatter or PickleFormatter()
        self.hasher = hasher or PickleHasher()
        self._function_registry: dict[str, FunctionInfo] = {}

    def __call__(self) -> Callable[[Callable[..., T]], Callable[..., T]]:
        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            funcinfo = FunctionInfo.build(func)
            self._function_registry[funcinfo.hash(self.hasher)] = funcinfo

            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> T:
                execinfo = ExecutionInfo.build(func, *args, **kwargs)
                key = ".".join((funcinfo.hash(self.hasher), execinfo.hash(self.hasher)))

                if self.storage.exists(key):
                    logger.info("Cache exists: %s", funcinfo.name)
                    with self.storage.open(key, self.formatter.READ_MODE) as file:
                        artifact = cast(T, self.formatter.read(file))
                else:
                    logger.info("Cache does not exists: %s", funcinfo.name)
                    artifact = func(*args, **kwargs)

                    logger.info("Store new artifact: %s", funcinfo.name)
                    with self.storage.open(key, self.formatter.WRITE_MODE) as file:
                        self.formatter.write(file, artifact)

                return artifact

            return wrapper

        return decorator

    def exists(self, func: Callable[..., Any] | FunctionInfo) -> bool:
        if not isinstance(func, FunctionInfo):
            func = FunctionInfo.build(func)
        prefix = func.hash(self.hasher)
        for key in self.storage.all():
            if key.startswith(prefix):
                return True
        return False

    def remove(self, func: Callable[..., Any] | FunctionInfo) -> None:
        if not isinstance(func, FunctionInfo):
            func = FunctionInfo.build(func)
        prefix = func.hash(self.hasher)
        for key in self.storage.all():
            if key.startswith(prefix):
                self.storage.remove(key)

    def funcinfos(self) -> list[FunctionInfo]:
        return list(self._function_registry.values())



import argparse
import importlib

from cachestore.commands.subcommand import Subcommand
from cachestore.common import Table
from cachestore.util import import_modules


@Subcommand.register("list")
class ListCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument("cache", nargs="+")
        self.parser.add_argument("--include-package", action="append", default=[])

    def run(self, args: argparse.Namespace) -> None:
        if args.include_package:
            import_modules(args.include_package)

        for cache_path in args.cache:
            cache_module, cache_name = cache_path.split(":", 1)
            cache = getattr(importlib.import_module(cache_module), cache_name)

            table = Table(columns=["name", "function", "cache"])
            for funcinfo in cache.funcinfos():
                table.add(
                    {
                        "name": cache_path,
                        "function": funcinfo.name,
                        "cache": "✓" if cache.exists(funcinfo) else "",
                    }
                )

        table.show()



import abc
from typing import IO, Any, ClassVar


class Formatter(abc.ABC):
    READ_MODE: ClassVar[str]
    WRITE_MODE: ClassVar[str]

    @abc.abstractmethod
    def write(self, file: IO[Any], obj: Any) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def read(self, file: IO[Any]) -> Any:
        raise NotImplementedError



import pickle
from typing import IO, Any, ClassVar

from cachestore.formatters.formatter import Formatter


class PickleFormatter(Formatter):
    READ_MODE: ClassVar = "rb"
    WRITE_MODE: ClassVar = "wb"

    def write(self, file: IO[Any], obj: Any) -> None:
        pickle.dump(obj, file)

    def read(self, file: IO[Any]) -> Any:
        return pickle.load(file)
